use crate::vulkan::vk_render_pass::VkRenderPass;
use crate::vulkan::{VkDevice, VkGraphicsPipeline, VkInstance, VkPhysicalDevice, VkSurface, VkSwapchain};
use ash::vk;
use std::sync::Arc;

pub struct VkRenderContext {
    instance: Arc<VkInstance>,
    surface: Arc<VkSurface>,
    device: Arc<VkDevice>,

    swapchain: Arc<VkSwapchain>,
}

impl VkRenderContext {
    pub fn init(window: &crate::display::Window) -> anyhow::Result<Self> {
        let required_extensions = window.required_extensions()?;

        let instance = Arc::new(VkInstance::new(&required_extensions));
        let surface = Arc::new(VkSurface::new(&window, instance.clone())?);

        let mut physical_devices = instance.get_physical_devices();
        physical_devices.sort_by(|a, b| b.priority().cmp(&a.priority()));

        let (physical_device, queue_family_index, properties) =
            VkPhysicalDevice::pick_physical_device_and_queue_by(
                &physical_devices,
                Some(vk::QueueFlags::GRAPHICS),
                Some(&surface),
            )
                .ok_or_else(|| anyhow::anyhow!("Unable to find physical device"))?;
        log::debug!(
            "Selected queue {properties:#?} for physical device {:?}",
            physical_device.properties.device_name_as_c_str()
        );

        let device = Arc::new(VkDevice::new_graphics_device(
            instance.clone(),
            &physical_device,
            queue_family_index,
        )?);

        let swapchain = Arc::new(VkSwapchain::new(
            &window,
            surface.clone(),
            device.clone(),
            &physical_device,
        )?);

        let render_pass = Arc::new(VkRenderPass::new(
            device.clone(),
            swapchain.clone(),
        )?);

        let pipeline = Arc::new(VkGraphicsPipeline::new(
            device.clone(),
            swapchain.clone(),
            render_pass.clone(),
        )?);

        let framebuffers = swapchain.present_image_views
            .as_ref()
            .ok_or_else(|| anyhow::anyhow!("No present image views found"))?
            .iter()
            .map(|present_image_view| {
                let attachments = [*present_image_view];
                let framebuffer_info = vk::FramebufferCreateInfo::default()
                    .render_pass(render_pass.handle)
                    .width(swapchain.surface_resolution.width)
                    .height(swapchain.surface_resolution.height)
                    .attachments(&attachments)
                    .layers(1);

                unsafe { device.handle.create_framebuffer(&framebuffer_info, None).unwrap() }
            })
            .collect::<Vec<_>>();

        let command_pool_info = vk::CommandPoolCreateInfo::default()
            .queue_family_index(device.queue_family_index);
        let command_pool = unsafe { device.handle.create_command_pool(&command_pool_info, None)? };

        let command_buffer_info = vk::CommandBufferAllocateInfo::default()
            .command_pool(command_pool)
            .level(vk::CommandBufferLevel::PRIMARY)
            .command_buffer_count(framebuffers.len() as u32);

        // Destroyed with command pool
        let command_buffers = unsafe { device.handle.allocate_command_buffers(&command_buffer_info)? };


        unsafe { device.handle.destroy_command_pool(command_pool, None) };
        for framebuffer in framebuffers {
            unsafe { device.handle.destroy_framebuffer(framebuffer, None) };
        }

        Ok(Self {
            instance,
            surface,
            device,

            swapchain,
        })
    }

    pub fn render(&mut self) -> anyhow::Result<()> {
        Ok(())
    }

    pub fn update_resolution(&mut self, width: u32, height: u32) -> anyhow::Result<()> {
        if let Some(swapchain) = Arc::get_mut(&mut self.swapchain) {
            swapchain.update_resolution(width, height)?;
        }

        Ok(())
    }
}
